#include <LagrangeFE.h>

int LagrangeFE:: nbf() {
  return StandardFE::nbf();       
}

void LagrangeFE:: compute_basis_functions() {

  // FIXME: in the below code dof(i) is not used to 
  // determine the basis functions 

  if (p->str() == "ReferenceLine") { 
    // Look at the case with the Triangle for a documented code
    ex polynom; 
    lst variables; 
  
    polynom = pol(order, 1, "a"); 
    variables = coeffs(polynom); 
  
    double increment = 1.0/order; 
  
    ex Nj; 
    for (int j=1; j <= order+1; j++) {
      lst equations; 
      int i=0; 
      for (double p=0; p<= 1.0 ; p += increment ) { 
        i++; 
        ex eq = polynom == dirac(i,j); 
        equations.append(eq.subs(x == p));  
        if (j == 1) dofs.insert(dofs.end(), p); 
      }
  
      ex subs = lsolve(equations, variables); 
      Nj = polynom.subs(subs);   
      Ns.insert(Ns.end(), Nj); 
    }
  

  } else if (p->str() == "ReferenceTriangle" ) { 
    // Look at the case with the Triangle for a documented code
    ex polynom; 
    lst variables; 
  
    polynom = pol(order, 2, "b"); 
    variables = coeffs(polynom); 
  
    double increment = 1.0/order; 
  
    ex Nj; 
    for (int j=1; j <= (order+1)*(order+2)/2; j++) {
      lst equations; 
      int i=0; 
      for (double p=0; p<= 1.0 ; p += increment ) { 
        for (double q=0; q<= 1.0-p ; q += increment ) { 
          i++; 
          ex eq = polynom == dirac(i,j); 
          equations.append(eq.subs(lst(x == p, y == q)));  
          if ( j == 1) dofs.insert(dofs.end(), lst(p,q)); //FIXME akward way to do it 
        }
      }
  
      ex subs = lsolve(equations, variables); 
      Nj = polynom.subs(subs);   
      Ns.insert(Ns.end(), Nj); 
    }
  } 
  else if ( p->str() == "Triangle" ){ 
    // Look HERE for the documented code
    ex polynom; 
    lst variables; 

    // the polynomial spaces on the form: 
    //      a0 + a1*x + a2*y + a3*x^2 + a4*x*y ...
    polynom = pol(order, 2, "b"); 
    // the variables a0,a1,a2 ..
    variables = coeffs(polynom); 
  
    ex Nj; 
    Polygon& pp = *p; 
    Triangle& t = (Triangle&) pp; 
    // The bezier ordinates in which the basis function should be either 0 or 1
    lst points = bezier_ordinates(t,order); 

    // Loop over all basis functions N_j and all points xi. 
    // Each basis function N_j is determined by a set of linear equations: 
    //   N_j(x_i) = dirac(i,j) 
    // This system of equations is then solved by lsolve
    for (int j=1; j <= points.nops(); j++) {
      lst equations; 
      int i=0; 
      for (int i=1; i<= points.nops() ; i++ ) { 
	// The point x_i 
	ex point = points.op(i-1); 
        // The equation N_j(x_i) = dirac(i,j)  
        ex eq = polynom == dirac(i,j); 
	// The equation is appended to the list of equations
        equations.append(eq.subs(lst(x == point.op(0) , y == point.op(1))));  
	// Creation of dof(j)
        if ( j == 1) dofs.insert(dofs.end(), lst(point.op(0),point.op(1))); //FIXME akward way to do it 
      }
  
      // We solve the linear system 
      ex subs = lsolve(equations, variables); 
      // Substitute to get the N_j 
      Nj = polynom.subs(subs);   
      // Append N_j to the list of basis functions 
      Ns.insert(Ns.end(), Nj); 

      // FIXME: In this case we create a linear system and solve it for each N_j.   
      // However, the matrix is always the same, it is only the right-hand side
      // that changes. Hence, it is possible to optimize here.  

    }

  } else if ( p->str() == "ReferenceTetrahedron" ) {
    // Look at the case with the Triangle for a documented code
    ex polynom; 
    lst variables; 
  
    polynom = pol(order, 3, "b"); 
    variables = coeffs(polynom); 
  
    int nno =0; 
    for (int j=0; j<= order; j++) { 
      nno += (j+1)*(j+2)/2; 
    }
  
    double increment = 1.0/order; 
  
    ex Nj; 
    for (int j=1; j <= nno; j++) {
      lst equations; 
      int i=0; 
      for (double p=0; p<= 1.0 ; p += increment ) { 
        for (double q=0; q<= 1.0-p ; q += increment ) { 
          for (double r=0; r<= 1.0-p-q ; r += increment ) { 
            i++; 
            ex eq = polynom == dirac(i,j); 
            equations.append(eq.subs(lst(x == p, y == q, z == r )));  
            if (j == 1) dofs.insert(dofs.end(), lst(p,q,r)); 
	  }
        }
      }
  
      ex subs = lsolve(equations, variables); 
      Nj = polynom.subs(subs);   
      Ns.insert(Ns.end(), Nj); 
    }
  }
  else if ( p->str() == "Tetrahedron" ){ 
    // Look at the case with the Triangle for a documented code
    ex polynom; 
    lst variables; 

  
    polynom = pol(order, 3, "b"); 
    variables = coeffs(polynom); 
  
    double increment = 1.0/order; 
  
    ex Nj; 
    Polygon& pp = *p; 
    Tetrahedron& t = (Tetrahedron&) pp; 
    lst points = bezier_ordinates(t,order); 
    for (int j=1; j <= points.nops(); j++) {
      lst equations; 
      int i=0; 
      for (int i=1; i<= points.nops() ; i++ ) { 
	ex point = points.op(i-1); 
        ex eq = polynom == dirac(i,j); 
        equations.append(eq.subs(lst(x == point.op(0) , y == point.op(1), z == point.op(2))));  
        if ( j == 1) dofs.insert(dofs.end(), lst(point.op(0),point.op(1),point.op(2))); //FIXME akward way to do it 
      }
  
      ex subs = lsolve(equations, variables); 
      Nj = polynom.subs(subs);   
      Ns.insert(Ns.end(), Nj); 
    }
  }
}



void LagrangeFE:: set(Polygon& p_) {
  StandardFE::set(p_); 
}


void LagrangeFE:: set(int order_) { 
  StandardFE::set(order_); 
}


ex LagrangeFE:: dof(int i) {
  return StandardFE::dof(i);  
}


ex LagrangeFE::N(int i) {  
  return StandardFE::N(i); 
}

// ------------VectorLagrangeFE --- 

int VectorLagrangeFE:: nbf() {
  return StandardFE::nbf();       
}

void VectorLagrangeFE:: compute_basis_functions() {
  LagrangeFE fe; 
  fe.set(order); 
  fe.set(*p); 
  fe.compute_basis_functions(); 
  lst zero_list; 
  for (int s=1; s<= size ; s++) {
    zero_list.append(0);  
  }

  for (int i=1; i<= fe.nbf() ; i++) {
    for (int s=1; s<= size ; s++) {
      lst Nis = zero_list;    
      Nis.let_op(s-1) = fe.N(i); 
      Ns.insert(Ns.end(), Nis);  

      lst dof = lst(fe.dof(i), s) ; 
      dofs.insert(dofs.end(), dof);  
    }
  }
}


void VectorLagrangeFE:: set_size(int size_) {
  size = size_; 
}


void VectorLagrangeFE:: set(Polygon& p_) {
  StandardFE::set(p_); 
}


void VectorLagrangeFE:: set(int order_) { 
  StandardFE::set(order_); 
}


ex VectorLagrangeFE:: dof(int i) {
  return StandardFE::dof(i);  
}


ex VectorLagrangeFE::N(int i) {  
  return StandardFE::N(i); 
}





