
#ifndef POLYGON_IS_INCLUDED
#define POLYGON_IS_INCLUDED

#include <tools.h>


class Polygon; 
class Triangle; 
class Mesh; 
class Point; 
class Edge; 
class Face; 

enum  Repr_format { 
  SUBS_PERFORMED = 1, 
  SUBS_NOT_PERFORMED = 2
}; 

/* Comment:  
   Every Polygon can have a name, e.g., for the lines in 
   a triangle we can name them   "l1", "l2", "l3".  
   In repr we could then represent the lines as
   line "l1": [x = A^{l1}*t -x0^{l1}, 
               y = B^{l1}*t -y0^{l1}, 
               z = C^{l1}*t -z0^{l1}, 
	       A^{l1} = .., B^{l1} = ..,  C^{l1} = .. 
	       t, 0, 1 ]

   can also consider to store this stuff as a map !  
   E.g., 
       equations = lst ( x = ..., y = ..., z = ...)   
       subst     = lst ( A^{li} = ..., B^{li} = ...,  C^{li} = ...)      
       variables = lst ( t ) 
       range     = lst ( 0, 1) 
*/ 

class Polygon {
  public: 
  string subscript; 
  exvector p; 

  Polygon() {}
  ~Polygon() {}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual ex integrate(ex f, Repr_format format = SUBS_PERFORMED) { 
	  //FIXME 
	  cout <<"Polygon integrate not implemented "<<endl; 
	  return ex(0); 
  }    
  virtual string str(); 
}; 

class Line : public Polygon  {
  ex a_; 
  ex b_; 
  public: 
  Line() {}
  Line(ex x0, ex x1, string subscript = ""); // x0_ and x1_ are points 
  ~Line(){}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual ex repr(ex t, Repr_format format = SUBS_PERFORMED); 
  virtual string str(); 
  virtual ex integrate(ex f, Repr_format format = SUBS_PERFORMED);    
}; 


class ReferenceLine : public Line {
  public: 
  ReferenceLine(string subscript = ""); 
  ~ReferenceLine(){}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual ex repr(ex t, Repr_format format = SUBS_PERFORMED); 
  ex a();  
  ex b(); 
  virtual string str(); 
  virtual ex integrate(ex f, Repr_format format = SUBS_PERFORMED); 
}; 

class Triangle : public Polygon {
  public: 
  Triangle(ex x0, ex x1, ex x1, string subscript = ""); 
  Triangle() {}
  ~Triangle(){}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual Line line(int i); 
  virtual ex repr(Repr_format = SUBS_PERFORMED); 
  virtual string str(); 
  virtual ex integrate(ex f, Repr_format format = SUBS_PERFORMED); 
}; 

class ReferenceTriangle : public Triangle {
  public: 
  ReferenceTriangle(string subscript = ""); 
  ~ReferenceTriangle(){}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual Line line(int i); 
  virtual ex repr(Repr_format format = SUBS_PERFORMED); 
  virtual string str(); 
  virtual ex integrate(ex f); 
}; 



class Tetrahedron : public Polygon {
  public: 
  Tetrahedron(string subscript) {}
  Tetrahedron(ex x0, ex x1, ex x1, ex x2, string subscript = ""); 
  ~Tetrahedron(){}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual Line line(int i); 
  virtual Triangle triangle(int i); 
  virtual ex repr(Repr_format format = SUBS_PERFORMED); 
  virtual string str(); 
  virtual ex integrate(ex f, Repr_format format = SUBS_PERFORMED);
}; 

class ReferenceTetrahedron : public Tetrahedron {
  public: 
  ReferenceTetrahedron(string subscript = ""); 
  ~ReferenceTetrahedron(){}

  virtual int no_vertices(); 
  virtual ex vertex(int i); 
  virtual Line line(int i); 
  virtual Triangle triangle(int i); 
  virtual ex repr(Repr_format format = SUBS_PERFORMED) { return ex(0); } 
  virtual string str(); 
  virtual ex integrate(ex f, Repr_format = SUBS_PERFORMED);  
}; 



// Some tools for Polygons 

// FIXME: change to barycenter(Triangle&)  
// FIXME:       and barycenter(Tetrahedron&)  
ex barycenter_triangle(ex p0, ex p1, ex p2); 
ex barycenter_tetrahedron(ex p0, ex p1, ex p2, ex p3); 
lst bezier_ordinates(Triangle& triangle, int d);  
lst bezier_ordinates(Tetrahedron& tetrahedra, int d);  




#endif

