#include <tools.h>

symbol x("x"); 
symbol y("y"); 
symbol z("z"); 
symbol infinity("infinity"); 
symbol DUMMY("DUMMY"); 

numeric nsd = 2; 


/*
void print(exvector& v) {
  for (int i=0; i< v.size(); i++) {
    cout <<"v["<<i<<"]="<<v[i]<<endl; 
  }
}
*/


	

void print(lst& l) {
//  for (lst::const_iterator i = l.begin(); i != l.end(); ++i)
//    cout << *i << endl;
//
  lst::const_iterator i = l.begin(); 
  cout <<"lst("<<*i; 
  ++i; 

  for (; i != l.end() ; ++i) {
    cout << ","<< *i;
  }
  cout <<");"<<endl; 
}

void print(exvector& v) {
  cout <<"v=["; 
  for (int i=0; i< v.size()-1; i++) {
    cout <<v[i]<<"," <<endl; 
  }
  cout <<v[v.size()-1]<< "]"<<endl; 

}

lst cross(lst& v1, lst& v2) {
  lst ret; 
  if ( v1.nops() != v2.nops() ) {
    cout <<"incompatible vectors "<<endl; 
    cout <<"v1.nops() "<<v1.nops(); 
    cout <<"  v2.nops() "<<v2.nops()<<endl; ; 
    return lst(); 
  }
  ret.append(  v1.op(1)*v2.op(2) - v1.op(2)*v2.op(1)); 
  ret.append(- v1.op(0)*v2.op(2) + v1.op(2)*v2.op(0)); 
  ret.append(  v1.op(0)*v2.op(1) - v1.op(1)*v2.op(0)); 
  return ret; 
}



ex inner(lst v1, lst v2) {
  ex ret; 
  if ( v1.nops() != v2.nops() ) {
    cout <<"incompatible vectors "<<endl; 
    cout <<"v1.nops() "<<v1.nops(); 
    cout <<"  v2.nops() "<<v2.nops()<<endl; ; 
    return 0; 
  }
  int i; 
  for (i = 0; i <= v1.nops()-1 ; ++i) {
	  ret += v1.op(i)*v2.op(i);  
       
  }
  return ret; 
}

lst matvec(matrix& M, lst& x) {
  lst ret; 
  int nr = M.rows(); 
  int nc = M.cols(); 
  for (int i = 0; i < nr; i++) {
    ex tmp; 
    for (int j = 0; j < nc; j++) {
       tmp = tmp +  M(i,j)*(x.op(j)); 
    }
    ret.append(tmp); 
  }
  return ret; 
}


ex inner(exvector& v1, exvector& v2){
  ex ret; 
  for (int i=0; i< v1.size(); i++) {
    ret += v1[i]*v2[i]; 
  }
  return ret; 
}



ex pol(int order, int nsd, const string a) {
  ex ret; // ex to return   
  int dof; // degrees of freedom  
  ex A;    // ex holding the coefficients a_0 .. a_dof  

  if (nsd == 1) {
    /* 1D: 
     * P^n = a_0 + a_1*x + .... + a_n*x^n 
     * dof : n+1
     */ 
    dof = order+1; 
    A = symbolic_matrix(1,dof, a); 
    int o=0; 
    for (const_iterator i = A.begin(); i != A.end(); ++i)  {  
      ret += (*i)*pow(x,o); 
      o++; 
    }
  }
  else if ( nsd == 2) {

    /* 2D: structure of coefficients (a_i)  
     * [ a_0      a_1 x     a_3 x^2     a_6 x^3  
     * [ a_2 y    a_4 xy    a_7 x^2y  
     * [ a_5 y^2  a_8 xy^2  
     * [ a_9 y^3 
     */
    dof = (order+1)*(order+2)/2; 
    A = symbolic_matrix(1, dof , a); 

    size_t i=0; 
    for (int o = 0; o <= order; o++) {
      for (int d = 0; d <= o; d++) { 
	ret += A.op(i)*pow(y,d)*pow(x,o-d); 
	i++; 
      }
    }
  }
  else if (nsd = 3) {

  /* Similar structure as in 2D, but 
   * structured as a tetraheder, i.e., 
   *   a_o + a_1 x + a_2 y + a_3 z 
   * + a_4 x^2 + a_5 xy +  
   */
    dof = 0; 
    for (int j=0; j<= order; j++) { 
      dof += (j+1)*(j+2)/2; 
    }
    A = symbolic_matrix(1, dof , a); 


    size_t i=0; 
    for (int o = 0; o <= order; o++) {
      for (int d = 0; d <= o; d++) { 
        for (int f = 0; f <= o; f++) { 
	  if ( o-d-f >= 0) {
            ret += A.op(i)*pow(y,f)*pow(z,d)*pow(x,o-d-f);  
            i++; 
	  }
	}
      }
    }
  }

  return ret; 
}

lst polv(int order, int nsd, const string a){
  lst ret; 
  for (int i=1; i<= nsd; i++) { 
    std::ostringstream s; 
    s <<a<<"^"<<i<<"_"; 
    ex p = pol(order, nsd, s.str()); 
    ret.append(p); 
  }
  return ret; 
}


ex polb(int order, int nsd, const string a) {

  ex ret; // ex to return   
  int dof; // degrees of freedom  
  ex A;    // ex holding the coefficients a_0 .. a_dof  

  if (nsd == 1) {
    /* 1D: 
     * P^n = a_0 + a_1*x + .... + a_n*x^n 
     * dof : n+1
     */ 
    dof = order+1; 
    A = symbolic_matrix(1,dof, a); 
    int o=0; 
    for (const_iterator i = A.begin(); i != A.end(); ++i)  {  
      ret += (*i)*pow(x,o); 
      o++; 
    }
  }
  else if ( nsd == 2) {

    /* 2D: structure of coefficients (a_i)  
     * [ a_0      a_1 x     a_3 x^2     a_6 x^3  
     * [ a_2 y    a_4 xy    a_7 x^2y  
     * [ a_5 y^2  a_8 xy^2  
     * [ a_9 y^3 
     */


    dof = (order+1)*(order+1); 
    A = symbolic_matrix(1, dof , a); 


    size_t i=0; 
    for (int o = 0; o <= order; o++) {
      for (int d = 0; d <= order; d++) { 
	ret += A.op(i)*pow(y,d)*pow(x,o); 
	i++; 
      }
    }
  }
  else if (nsd = 3) {

  /* Similar structure as in 2D, but 
   * structured as a tetraheder, i.e., 
   *   a_o + a_1 x + a_2 y + a_3 z 
   * + a_4 x^2 + a_5 xy +  
   */
    dof = (order+1)*(order+1)*(order+1); 
    A = symbolic_matrix(1, dof , a); 


    size_t i=0; 
    for (int o = 0; o <= order; o++) {
      for (int d = 0; d <= order; d++) { 
        for (int f = 0; f <= order; f++) { 
            ret += A.op(i)*pow(y,f)*pow(z,d)*pow(x,o);  
            i++; 
	}
      }
    }
  }

  return ret; 
}

ex div(lst& v) {
  nsd = v.nops();  
  ex ret; 
  if (nsd == 1) {
    ret = v.op(0).diff(x);
  }
  else if (nsd == 2) {
    ret = v.op(0).diff(x) + v.op(1).diff(y); 
  }
  else if (nsd == 3) {
    ret = v.op(0).diff(x) + v.op(1).diff(y) + v.op(2).diff(z); 
  }
  return ret; 
}

ex div(exvector& v) {
  ex ret; 
  if (nsd == 2) {
    ret = v[0].diff(x) + v[1].diff(y); 
  }
  else if (nsd == 3) {
    ret = v[0].diff(x) + v[1].diff(y) + v[2].diff(z); 
  }
  return ret; 
}

lst coeffs(lst pols) {
  lst cc; 
  lst tmp; 
  for (int i=0; i<= pols.nops()-1; i++) {
    tmp = coeffs(pols.op(i)); 
    cc = collapse(lst(cc, tmp)); 
  }
  return cc; 
}

lst coeffs(ex pol) {
  lst cc; 
  ex c, b; 
  for (int i=pol.ldegree(x); i<=pol.degree(x); ++i) {
    for (int j=pol.ldegree(y); j<=pol.degree(y); ++j) {
      for (int k=pol.ldegree(z); k<=pol.degree(z); ++k) {
        c = pol.coeff(x,i).coeff(y, j).coeff(z,k); 
        if ( c != 0 ) cc.append(c); 
      }
    }
  }
  return cc; 
}



exvector coeff(ex pol) {
  exvector cc; 
  ex c, b; 
  for (int i=pol.ldegree(x); i<=pol.degree(x); ++i) {
    for (int j=pol.ldegree(y); j<=pol.degree(y); ++j) {
      for (int k=pol.ldegree(z); k<=pol.degree(z); ++k) {
        c = pol.coeff(x,i).coeff(y, j).coeff(z,k); 
        if ( c != 0 ) cc.insert(cc.begin(),c); 
      }
    }
  }
  return cc; 
}

int dirac(int i, int j) {
  if (i==j) return 1; 
  else return 0; 
}

ex_ex_map pol2basisandcoeff(ex e) { 
  ex c; 
  ex b; 
  ex_ex_map map; 
  for (int i=e.ldegree(x); i<=e.degree(x); ++i) {
    for (int j=e.ldegree(y); j<=e.degree(y); ++j) {
      for (int k=e.ldegree(z); k<=e.degree(z); ++k) {
        c = e.coeff(x,i).coeff(y, j).coeff(z,k); 
        b = pow(x,i)*pow(y,j)*pow(z,k); 
        map[b] = c;  
      }
    }
  }
  return map; 
}

void print(ex_int_map map) {
  ex b; 
  int c; 
  ex_int_it iter; 
  iter = map.begin(); 
  cout <<"{" <<b<<":"<<c; 
  for (iter = map.begin(); iter != map.end(); iter++) {  
    b = (*iter).first; c = map[b]; 
    cout <<", "<<b<<":"<<c; 
  }
  cout <<"}"<<endl; 

}



void print(ex_ex_map map) {
  ex b; 
  ex c; 
  ex_ex_it iter; 
  iter = map.begin(); 
  cout <<"{" <<b<<":"<<c; 
  for (iter = map.begin(); iter != map.end(); iter++) {  
    b = (*iter).first; c = map[b]; 
    cout <<", "<<b<<":"<<c; 
  }
  cout <<"}"<<endl; 

}


lst ex2equations(ex rel) {
  ex lhs = rel.lhs();  
  ex rhs = rel.rhs(); 

  ex l; 
  ex r; 

  lst eqs; 

  for (int i=lhs.ldegree(x); i<=lhs.degree(x); ++i) {
    for (int j=lhs.ldegree(y); j<=lhs.degree(y); ++j) {
      for (int k=lhs.ldegree(z); k<=lhs.degree(z); ++k) {
        l = lhs.coeff(x,i).coeff(y, j).coeff(z,k); 
        r = rhs.coeff(x,i).coeff(y, j).coeff(z,k); 
//	if (! (l == 0 && r == 0 ) )  eqs.append(l == r); OLD VERSION 
	if ( (l != 0 && (r == 0 || r == 1) ) )  eqs.append(l == r); 
      }
    }
  }
  eqs.sort(); 
  return eqs; 
}

lst collapse(lst l) {
  lst lc;  
  lst::const_iterator iter1, iter2; 

  for (iter1 = l.begin(); iter1 != l.end(); ++iter1) {
     if (is_a<lst>(*iter1)) {
       for (iter2 = ex_to<lst>(*iter1).begin(); iter2 != ex_to<lst>(*iter1).end(); ++iter2) {
          lc.append(*iter2); 
       }
     } else {
       lc.append(*iter1); 
     }
  }
  lc.sort(); 
  lc.unique(); 
return lc; 
}


matrix equations2matrix(const ex &eqns, const ex &symbols) {

  cout <<"no equations "<<eqns.nops()<<endl; 
  cout <<"no variables "<<symbols.nops()<<endl; 
 	
  matrix sys(eqns.nops(),symbols.nops());
  matrix rhs(eqns.nops(),1);
  matrix vars(symbols.nops(),1);
	
  for (size_t r=0; r<eqns.nops(); r++) {
    const ex eq = eqns.op(r).op(0)-eqns.op(r).op(1); // lhs-rhs==0
    ex linpart = eq;
    for (size_t c=0; c<symbols.nops(); c++) {
      const ex co = eq.coeff(ex_to<symbol>(symbols.op(c)),1);
      linpart -= co*symbols.op(c);
      sys(r,c) = co;
    }
    linpart = linpart.expand();
    rhs(r,0) = -linpart;
  }
  cout <<"matrix "<<sys<<endl; 
  cout <<"rhs "<<rhs<<endl; 
  cout <<"vars"<<vars<<endl; 
  return sys; 
}

lst grad(ex f) {
  if (nsd == 1) {
    return lst(diff(f,x)); 
  } else if ( nsd == 2) {
    return lst(diff(f,x), diff(f,y));  
  } else if ( nsd == 3) {
    return lst(diff(f,x), diff(f,y), diff(f,z));  
  } else {
    //FIXME: proper error handling
    cout <<"THIS SHOULD NEVER HAPPEN"; 
    return lst(); 
  }

}

lst lst_equals(ex a, ex b) { 
  lst ret; 
  if ( (is_a<lst>(a)) && (is_a<lst>(b)) /*&& (a.nops() == b.nops())*/ ) { 
    for (int i=0; i<= a.nops()-1; i++) {
      ret.append(b.op(i) == a.op(i)); 
    }
  } else if ( !(is_a<lst>(a)) && !(is_a<lst>(b))) { 
      ret.append(b == a); 
  } else if ( !(is_a<lst>(a)) && (is_a<lst>(b))) { 
      ret.append(b.op(0) == a); 
  } else {
      //FIXME: proper error handling
      cout <<"THIS SHOULD NEVER HAPPEN"<<endl;    
  }
  return ret; 
}

ex lst_to_matrix2(const lst& l)
{
     lst::const_iterator itr, itc;
 
     // Find number of rows and columns
     size_t rows = l.nops(), cols = 0;
     for (itr = l.begin(); itr != l.end(); ++itr) {
         if (!is_a<lst>(*itr))
//              throw (std::invalid_argument("lst_to_matrix: argument must be a list of lists"));
             cols = 1; 
         if (itr->nops() > cols)
             cols = itr->nops();
     }
     // Allocate and fill matrix
     matrix &M = *new matrix(rows, cols);
     M.setflag(status_flags::dynallocated);
 
     unsigned i;
     for (itr = l.begin(), i = 0; itr != l.end(); ++itr, ++i) {
         unsigned j;
	 if (cols == 1) {
             M(i, 0) = *itr;
	 } else {
           for (itc = ex_to<lst>(*itr).begin(), j = 0; itc != ex_to<lst>(*itr).end(); ++itc, ++j)
               M(i, j) = *itc;
	 }
     }
     return M;
}


lst matrix_to_lst2(const ex& m) {
   //FIXME : should check that this is a matrix 
   if (is_a<matrix>(m)) {
     matrix A = ex_to<matrix>(m);  
     int cols = A.cols(); 
     int rows = A.rows(); 
  
  
     lst ret; 
     if ( cols == 1) {
       for (int i=0; i<=A.rows()-1; i++) { 
         ret.append(A(i,0)); 
       }
     } else {
       for (int i=0; i<=A.rows()-1; i++) { 
         lst rl; 
         for (int j=0; j<=A.cols()-1; j++) { 
           rl.append(A(i,j)); 
         }
         ret.append(rl); 
       }
     }
     return ret; 
   } else { 
     return lst(); 
   }

}



int find(ex e, lst list){
  for (int i=0; i< list.nops(); i++) {
    if ( e == list.op(i) ) return i; 
  }
  return -1; 
}

void visitor_subst_pow(ex e, ex_ex_map& map, ex_int_map& intmap, string a) {  
  static int i=0;  
  if (map.find(e) != map.end()) { 
    intmap[e] = intmap[e]+1;
    return;   
  }
  if (is_exactly_a<power>(e)) { 
    std::ostringstream s; 
    s <<a<<i++; 
    map[e] = symbol(s.str()); 
    intmap[e] = 0;  
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"power e "<<e2<<endl; 
       visitor_subst_pow(e2,map,intmap, a); 
    }
  }
  else if (is_a<function>(e)) { 
    std::ostringstream s; 
    s <<a<<i++; 
    map[e] = symbol(s.str()); 
    intmap[e] = 0;  
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"function e "<<e2<<endl; 
       visitor_subst_pow(e2,map,intmap, a); 
    }
  }
  else if (is_a<mul>(e)) { 
    if (e.nops() > 4 && e.nops() < 10 ) { 
      std::ostringstream s; 
      s <<a<<i++; 
      map[e] = symbol(s.str()); 
      intmap[e] = 0;  
    }

    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
       visitor_subst_pow(e2,map,intmap, a); 
    }
  }
  else if (is_a<add>(e)) { 
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
       visitor_subst_pow(e2,map,intmap,a); 
    }
  }


}




void check_visitor(ex e, lst& exlist) {
  if (find(e, exlist) >= 0) return;   

//  cout <<"ex e "<<e<<endl; 
  if (is_a<numeric>(e)) { 
  }
  else if (is_a<add>(e) ) {
//    cout <<"e "<<e <<endl; 
//    cout <<"e.nops() "<<e.nops() <<endl; 
    if (e.nops() > 4 && e.nops() < 10 ) exlist.append(e); 
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"add e "<<e2<<endl; 
//       exlist.append(e2); 
       check_visitor(e2,exlist); 
    }
  } 
  else if (is_a<mul>(e)) { 
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"mul e "<<e2<<endl; 
       exlist.append(e2); 
       check_visitor(e2,exlist); 
    }
  }
  else if (is_a<lst>(e)) { 
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"lst e "<<e2<<endl; 
//       exlist.append(e2); 
       check_visitor(e2,exlist); 
    }
  }
  else if (is_exactly_a<power>(e)) { 
    exlist.append(e); 
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"power e "<<e2<<endl; 
       check_visitor(e2,exlist); 
    }
  }
  else if (is_a<function>(e)) { 
    exlist.append(e); 
    for (int i=0; i< e.nops(); i++) { 
       ex e2 = e.op(i);  
//       cout <<"function e "<<e2<<endl; 
       check_visitor(e2,exlist); 
    }
  }



  else {
//       exlist.append(e); 
//    cout <<"atom e "<<e<<endl; 
  }

  exlist.sort(); 
  exlist.unique(); 
}


string istr(string a, int b) {
  std::ostringstream s; 
  s <<a<<b; 
  return s.str(); 
}

void EQUAL_OR_DIE(ex e, char* s) {
  if (!compare(e, string(s))) { 
    cout <<"ERROR: expression e: " <<e<<" is not equal to "<<s<<endl; 
    exit(-1); 
  }
}


bool compare(ex e, string s) { 
  std::ostringstream ss; 
  ss<<e; 
  string es = ss.str(); 
  if ( es == s) return true;  
  else return false; 
}

