#ifndef TOOLS_IS_INCLUDED 
#define TOOLS_IS_INCLUDED 

#include <ginac/ginac.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <sstream>
#include <algorithm>
#include <map>
#include <stdexcept>
#include <vector>



using namespace GiNaC; 
using namespace std; 

extern symbol x; 
extern symbol y; 
extern symbol z; 
extern numeric nsd; 
extern symbol infinity; 
extern symbol DUMMY; 

/* Vector of ex */ 
typedef vector<ex> ex_vector; 

/* Dictionary with ex as keys and values */ 
typedef map<ex,ex, ex_is_less>  ex_ex_map;
typedef map<ex,int, ex_is_less>  ex_int_map;

/* Iterator associated with the ex-ex dictionary */ 
typedef map<ex,ex>::iterator ex_ex_it;  
typedef map<ex,int>::iterator ex_int_it;  

/* Vector of vectors of ex */ 
typedef vector<exvector> ex_vector_vector;


// inner product of vectors or lst
ex inner(exvector& v1, exvector& v2); 
ex inner(lst v1, lst v2); 
lst cross(lst& v1, lst& v2); 


// generates a polynom of arbitrary order on a line, a triangle, or a tetrahedron 
ex pol(int order, int nsd, const string a);  
// generates a vector polynom of arbitrary order on a line, a triangle or a tetrahedron 
lst polv(int order, int nsd, const string a);  
// generates a polynom of arbitrary order on a square or a box   
ex polb(int order, int nsd, const string a);  
// generates a vector polynom of arbitrary order on a squart or a box  
lst polbv(int order, int nsd, const string a);  


// matrix vector product
lst matvec(matrix& M, lst& x);  



// extracts the coefficents from a polynomial
exvector coeff(ex pol); 
lst coeffs(ex pol); 
lst coeffs(lst pols); 
// extract the basisfunctions and corresponding coefficents from a polynomial
ex_ex_map  pol2basisandcoeff(ex e); 


// the divergence 
ex div(exvector& v);  
ex div(lst& v);  
lst grad(ex f); 

// print functions 
void print(exvector& v); 
void print(lst& l); 
void print(ex_ex_map m); 
void print(ex_int_map m); 

// dirac delta function 
int dirac(int i, int j); 


lst ex2equations(ex rel); 
lst collapse(lst l);  

matrix equations2matrix (const ex &eqns, const ex &symbols); 

lst lst_equals(ex a, ex b);  
ex lst_to_matrix2(const lst & l); 
lst matrix_to_lst2(const ex & m ); 

// FIXME bad name
int find(ex e, lst list); 
void check_visitor(ex e, lst& exlist);  
void visitor_subst_pow(ex e, ex_ex_map& map, ex_int_map& intmap, string a);  

string istr(string a, int b); 

bool compare(ex e, string s); 

void EQUAL_OR_DIE(ex e, char* s);  


#endif 




